var EventEmitterModule = require("events");
var EventEmitter = EventEmitterModule.EventEmitter;
var Promise = require("q");
var zookeeper = require("node-zookeeper-client");
var Exception = zookeeper.Exception;
var LocatorException = require("./locatorException");
var ClientWrapper = (function () {
    function ClientWrapper() {
        this.emitter = new EventEmitter();
        this.client = null;
    }
    ClientWrapper.prototype.setClient = function (client) {
        this.client = client;
        this.emitter.emit('NEW_CLIENT');
    };
    return ClientWrapper;
})();
function defaultDataExtractor(data) {
    var zkJS;
    try {
        zkJS = JSON.parse(data);
    }
    catch (e) {
        return null;
    }
    if (!zkJS.address || !zkJS.port) {
        return null;
    }
    return {
        host: zkJS.address,
        port: zkJS.port
    };
}
function defaultServerExtractor(data) {
    return JSON.parse(data);
}
function makeManagerForPath(clientWrapper, path, emitter, dataExtractor, locatorTimeout, strict) {
    var next = -1;
    var pool = null;
    var cachedPool = null;
    var queue = [];
    var stale = false;
    function setPool(newPool) {
        pool = newPool;
        if (!cachedPool || (newPool && newPool.length)) {
            cachedPool = newPool;
        }
        return;
    }
    function getPool() {
        if (strict || (pool && pool.length)) {
            return pool;
        }
        if (cachedPool && cachedPool.length) {
            return cachedPool;
        }
        return pool;
    }
    function dispatch(deferred) {
        var chosenPool = getPool();
        if (!chosenPool) {
            throw new Error("get next called on loading pool");
        }
        if (chosenPool.length) {
            next++;
            return deferred.resolve(chosenPool[next % chosenPool.length]);
        }
        else {
            return deferred.reject(new Error(LocatorException.CODE["EMPTY_POOL"]));
        }
    }
    function processQueue() {
        while (queue.length) {
            dispatch(queue.shift());
        }
    }
    function onGetChildren(err, children) {
        if (err) {
            emitter.emit(LocatorException.CODE["FAILED_TO_GET_CHILDREN"], path, err);
            setPool([]);
            processQueue();
            return;
        }
        var promises = children.map(function (child) {
            var deferred = Promise.defer();
            clientWrapper.client.getData(path + "/" + child, function (_err, data) {
                if (_err) {
                    if (_err.getCode() !== Exception.NO_NODE) {
                        emitter.emit(LocatorException.CODE["FAILED_TO_GET_CHILD_INFO"], path, _err);
                    }
                    deferred.resolve(null);
                    return;
                }
                deferred.resolve(dataExtractor(data.toString("utf8")));
                return;
            });
            return deferred.promise;
        });
        Promise.all(promises)
            .then(function (newPool) {
            setPool(newPool.filter(Boolean));
            emitter.emit(LocatorException.CODE["NEW_POOL"], path, pool);
            processQueue();
        })
            .done();
    }
    function onChildrenChange(event) {
        emitter.emit(LocatorException.CODE["CHILDREN_CHANGED"], path, event);
        clientWrapper.client.getChildren(path, onChildrenChange, onGetChildren);
    }
    function onExists(error, stat) {
        if (stat) {
            stale = false;
            emitter.emit(LocatorException.CODE["PATH_FOUND"], path);
            clientWrapper.client.getChildren(path, onChildrenChange, onGetChildren);
        }
        else {
            stale = true;
            emitter.emit(LocatorException.CODE["PATH_NOT_FOUND"], path);
            setPool([]);
            processQueue();
        }
    }
    clientWrapper.emitter.on('NEW_CLIENT', function () {
        setPool(null);
        clientWrapper.client.exists(path, onExists);
    });
    if (clientWrapper.client) {
        clientWrapper.client.exists(path, onExists);
    }
    return function () {
        var deferred = Promise.defer();
        if (stale) {
            setPool(null);
            clientWrapper.client.exists(path, onExists);
        }
        if (pool) {
            dispatch(deferred);
        }
        else {
            queue.push(deferred);
        }
        if (locatorTimeout) {
            setTimeout((function () {
                if (deferred.promise.isPending()) {
                    deferred.reject(new Error("ZOOKEEPER_TIMEOUT"));
                }
            }), locatorTimeout);
        }
        return deferred.promise;
    };
}
function zookeeperLocatorFactory(parameters) {
    var serverLocator = parameters.serverLocator;
    var path = parameters.path;
    var dataExtractor = parameters.dataExtractor;
    var locatorTimeout = parameters.locatorTimeout;
    var sessionTimeout = parameters.sessionTimeout;
    var spinDelay = parameters.spinDelay;
    var retries = parameters.retries;
    var strict = parameters.strict;
    dataExtractor = dataExtractor ? dataExtractor : defaultDataExtractor;
    locatorTimeout = locatorTimeout ? locatorTimeout : 2000;
    sessionTimeout = sessionTimeout ? sessionTimeout : 10000;
    retries = retries ? retries : 0;
    if (!(typeof strict !== 'undefined' && strict !== null)) {
        strict = true;
    }
    var emitter = new EventEmitter();
    var client;
    var clientWrapper = new ClientWrapper();
    var connect = function () {
        serverLocator()
            .then(function (location) {
            var zookeeperServer = location.host;
            if (location.port) {
                zookeeperServer = zookeeperServer + ":" + location.port;
            }
            zookeeperServer = zookeeperServer + path;
            client = zookeeper.createClient(zookeeperServer, {
                sessionTimeout: sessionTimeout,
                spinDelay: spinDelay,
                retries: retries
            });
            clientWrapper.setClient(client);
            var disconnectedHandler = null;
            client.on("connected", function () {
                emitter.emit(LocatorException.CODE["CONNECTED"]);
                if (disconnectedHandler) {
                    clearTimeout(disconnectedHandler);
                }
            });
            client.on("disconnected", function (count) {
                emitter.emit(LocatorException.CODE["DISCONNECTED"]);
                disconnectedHandler = setTimeout(function () {
                    client.close();
                    setTimeout(function () {
                        client.emit("expired");
                    }, 1000);
                }, sessionTimeout);
            });
            client.on('state', function (state) { return emitter.emit(LocatorException.CODE["STATE_CHANGE"], state, client.getSessionId()); });
            client.on("expired", function () {
                emitter.emit(LocatorException.CODE["EXPIRED"]);
                connect();
            });
            emitter.emit(LocatorException.CODE["CONNECTING"]);
            client.connect();
        })
            .catch(function (err) {
            emitter.emit(LocatorException.CODE["ZK_LOCATOR_ERROR"], err);
            connect();
        })
            .done();
    };
    connect();
    var pathManager = {};
    function manager(_path) {
        if (typeof _path !== "string") {
            throw new TypeError("path must be a string");
        }
        if (_path[0] !== "/") {
            _path = "/" + _path;
        }
        pathManager[_path] = pathManager[_path] ? pathManager[_path] : makeManagerForPath(clientWrapper, _path, emitter, dataExtractor, locatorTimeout, strict);
        return pathManager[_path];
    }
    Object.keys(EventEmitter.prototype).forEach(function (fnName) {
        return manager[fnName] = function () {
            return emitter[fnName].apply(emitter, arguments);
        };
    });
    return manager;
}
exports.zookeeperLocatorFactory = zookeeperLocatorFactory;
